# flake8: noqa
# mypy: ignore-errors
from functools import lru_cache
from typing import Dict, List, Optional

from nautilus_trader.adapters.binance.common.enums import BinanceAccountType
from nautilus_trader.adapters.binance.config import (
    BinanceExecClientConfig as NautilusBinanceExecClientConfig,
)
from nautilus_trader.adapters.binance.factories import _get_http_base_url
from nautilus_trader.adapters.binance.factories import _get_ws_base_url
from nautilus_trader.adapters.binance.factories import get_cached_binance_http_client
from nautilus_trader.adapters.binance.futures import data as futures_data
from nautilus_trader.adapters.binance.futures import execution as futures_execution
from nautilus_trader.adapters.binance.futures import providers as futures_providers
from nautilus_trader.adapters.binance.futures.http.account import BinanceFuturesAccountHttpAPI
from nautilus_trader.adapters.binance.futures.http.market import BinanceFuturesMarketHttpAPI
from nautilus_trader.adapters.binance.futures.http.user import BinanceFuturesUserDataHttpAPI
from nautilus_trader.adapters.binance.futures.http.wallet import BinanceFuturesWalletHttpAPI
from nautilus_trader.adapters.binance.http.error import BinanceClientError
from nautilus_trader.adapters.binance.spot import data as spot_data
from nautilus_trader.adapters.binance.spot import execution as spot_execution
from nautilus_trader.adapters.binance.spot import providers as spot_providers
from nautilus_trader.adapters.binance.spot.http.account import BinanceSpotAccountHttpAPI
from nautilus_trader.adapters.binance.spot.http.market import BinanceSpotMarketHttpAPI
from nautilus_trader.adapters.binance.spot.http.user import BinanceSpotUserDataHttpAPI
from nautilus_trader.adapters.binance.spot.http.wallet import BinanceSpotWalletHttpAPI
from nautilus_trader.adapters.binance.websocket.client import BinanceWebSocketClient
from nautilus_trader.common.logging import LogColor
from nautilus_trader.core.correctness import PyCondition
from nautilus_trader.core.datetime import millis_to_nanos
from nautilus_trader.live.factories import LiveDataClientFactory
from nautilus_trader.live.factories import LiveExecClientFactory
from nautilus_trader.model.enums import AccountType
from nautilus_trader.model.enums import OMSType
from nautilus_trader.model.identifiers import AccountId
from nautilus_trader.model.identifiers import ClientId
from nautilus_trader.model.identifiers import InstrumentId
from nautilus_trader.model.identifiers import Venue

from nacre.live.execution_client import LiveExecutionClient


# All the shitty code below patch custom venue/account_id/oms_type into clients
# Original nautilus clients locked venue to BINANCE_VENUE
# which not work in situations like some spot-futures arbitrage strategies that require both account type

# Temporary solution. Consider create PR to nautilus_trader


class BinanceExecClientConfig(NautilusBinanceExecClientConfig):
    oms_type: OMSType = OMSType.NETTING


@lru_cache(1)
def get_cached_binance_spot_instrument_provider(
    client,
    logger,
    venue,
    account_type,
    config,
):
    return spot_providers.BinanceSpotInstrumentProvider(
        client=client,
        logger=logger,
        venue=venue,
        account_type=account_type,
        config=config,
    )


@lru_cache(1)
def get_cached_binance_futures_instrument_provider(
    client,
    logger,
    venue,
    account_type,
    config,
):
    return futures_providers.BinanceFuturesInstrumentProvider(
        client=client,
        logger=logger,
        venue=venue,
        account_type=account_type,
        config=config,
    )


def spot_init(
    self,
    loop,
    client,
    msgbus,
    cache,
    clock,
    logger,
    instrument_provider,
    name,
    account_id,
    venue,
    account_type: BinanceAccountType = BinanceAccountType.SPOT,
    base_url_ws=None,
):
    super(spot_execution.BinanceSpotExecutionClient, self).__init__(
        loop=loop,
        client_id=ClientId(name),
        venue=venue,
        oms_type=OMSType.NETTING,
        instrument_provider=instrument_provider,
        account_type=AccountType.CASH,
        base_currency=None,
        msgbus=msgbus,
        cache=cache,
        clock=clock,
        logger=logger,
    )

    self._binance_account_type = account_type
    self._log.info(f"Account type: {self._binance_account_type.value}.", LogColor.BLUE)

    self._set_account_id(account_id)

    # HTTP API
    self._http_client = client
    self._http_account = BinanceSpotAccountHttpAPI(client=client)
    self._http_market = BinanceSpotMarketHttpAPI(client=client)
    self._http_user = BinanceSpotUserDataHttpAPI(client=client, account_type=account_type)

    # Listen keys
    self._ping_listen_keys_interval = 60 * 5  # Once every 5 mins (hardcode)
    self._ping_listen_keys_task = None
    self._listen_key = None

    # WebSocket API
    self._ws_client = BinanceWebSocketClient(
        loop=loop,
        clock=clock,
        logger=logger,
        handler=self._handle_user_ws_message,
        base_url=base_url_ws,
    )

    # Hot caches
    self._instrument_ids = {}

    self._log.info(f"Base URL HTTP {self._http_client.base_url}.", LogColor.BLUE)
    self._log.info(f"Base URL WebSocket {base_url_ws}.", LogColor.BLUE)


def futures_init(
    self,
    loop,
    client,
    msgbus,
    cache,
    clock,
    logger,
    instrument_provider,
    name,
    account_id,
    venue,
    oms_type,
    account_type: BinanceAccountType = BinanceAccountType.FUTURES_USDT,
    base_url_ws=None,
):
    super(futures_execution.BinanceFuturesExecutionClient, self).__init__(
        loop=loop,
        client_id=ClientId(name),
        venue=venue,
        oms_type=oms_type,
        instrument_provider=instrument_provider,
        account_type=AccountType.MARGIN,
        base_currency=None,
        msgbus=msgbus,
        cache=cache,
        clock=clock,
        logger=logger,
    )

    self._binance_account_type = account_type
    self._log.info(f"Account type: {self._binance_account_type.value}.", LogColor.BLUE)

    self._set_account_id(account_id)
    self._set_venue(venue)

    # HTTP API
    self._http_client = client
    self._http_account = BinanceFuturesAccountHttpAPI(client=client, account_type=account_type)
    self._http_market = BinanceFuturesMarketHttpAPI(client=client, account_type=account_type)
    self._http_user = BinanceFuturesUserDataHttpAPI(client=client, account_type=account_type)

    # Listen keys
    self._ping_listen_keys_interval = 60 * 5  # Once every 5 mins (hardcode)
    self._ping_listen_keys_task = None
    self._listen_key = None

    # WebSocket API
    self._ws_client = BinanceWebSocketClient(
        loop=loop,
        clock=clock,
        logger=logger,
        handler=self._handle_user_ws_message,
        base_url=base_url_ws,
    )

    # Hot caches
    self._instrument_ids = {}

    self._log.info(f"Base URL HTTP {self._http_client.base_url}.", LogColor.BLUE)
    self._log.info(f"Base URL WebSocket {base_url_ws}.", LogColor.BLUE)


spot_execution.BinanceSpotExecutionClient = type(
    "BinanceSpotExecutionClient",
    (LiveExecutionClient,),
    dict(spot_execution.BinanceSpotExecutionClient.__dict__),
)
futures_execution.BinanceFuturesExecutionClient = type(
    "BinanceFuturesExecutionClient",
    (LiveExecutionClient,),
    dict(futures_execution.BinanceFuturesExecutionClient.__dict__),
)
spot_execution.BinanceSpotExecutionClient.__init__ = spot_init
futures_execution.BinanceFuturesExecutionClient.__init__ = futures_init


def spot_data_init(
    self,
    loop,
    client,
    msgbus,
    cache,
    clock,
    logger,
    instrument_provider,
    venue,
    account_type: BinanceAccountType = BinanceAccountType.SPOT,
    base_url_ws=None,
):
    super(spot_data.BinanceSpotDataClient, self).__init__(
        loop=loop,
        client_id=ClientId(venue.value),
        venue=venue,
        instrument_provider=instrument_provider,
        msgbus=msgbus,
        cache=cache,
        clock=clock,
        logger=logger,
    )

    assert account_type.is_spot or account_type.is_margin, "account type is not for spot/margin"
    self._binance_account_type = account_type
    self._log.info(f"Account type: {self._binance_account_type.value}.", LogColor.BLUE)

    self._update_instrument_interval = 60 * 60  # Once per hour (hardcode)
    self._update_instruments_task = None

    # HTTP API
    self._http_client = client
    self._http_market = BinanceSpotMarketHttpAPI(client=self._http_client)  # type: ignore

    # WebSocket API
    self._ws_client = BinanceWebSocketClient(
        loop=loop,
        clock=clock,
        logger=logger,
        handler=self._handle_ws_message,
        base_url=base_url_ws,
    )

    # Hot caches
    self._instrument_ids = {}
    self._book_buffer = {}

    self._log.info(f"Base URL HTTP {self._http_client.base_url}.", LogColor.BLUE)
    self._log.info(f"Base URL WebSocket {base_url_ws}.", LogColor.BLUE)


def futures_data_init(
    self,
    loop,
    client,
    msgbus,
    cache,
    clock,
    logger,
    instrument_provider,
    venue,
    account_type: BinanceAccountType = BinanceAccountType.FUTURES_USDT,
    base_url_ws=None,
):
    super(futures_data.BinanceFuturesDataClient, self).__init__(
        loop=loop,
        client_id=ClientId(venue.value),
        venue=venue,
        instrument_provider=instrument_provider,
        msgbus=msgbus,
        cache=cache,
        clock=clock,
        logger=logger,
    )

    assert account_type.is_futures, "account type is not for futures"
    self._binance_account_type = account_type
    self._log.info(f"Account type: {self._binance_account_type.value}.", LogColor.BLUE)

    self._update_instrument_interval = 60 * 60  # Once per hour (hardcode)
    self._update_instruments_task = None

    # HTTP API
    self._http_client = client
    self._http_market = BinanceFuturesMarketHttpAPI(client=client, account_type=account_type)
    self._http_user = BinanceFuturesUserDataHttpAPI(client=client, account_type=account_type)

    # Listen keys
    self._ping_listen_keys_interval = 60 * 5  # Once every 5 mins (hardcode)
    self._ping_listen_keys_task = None
    self._listen_key = None

    # WebSocket API
    self._ws_client = BinanceWebSocketClient(
        loop=loop,
        clock=clock,
        logger=logger,
        handler=self._handle_ws_message,
        base_url=base_url_ws,
    )

    # Hot caches
    self._instrument_ids = {}
    self._book_buffer = {}

    self._log.info(f"Base URL HTTP {self._http_client.base_url}.", LogColor.BLUE)
    self._log.info(f"Base URL WebSocket {base_url_ws}.", LogColor.BLUE)


spot_data.BinanceSpotDataClient.__init__ = spot_data_init
futures_data.BinanceFuturesDataClient.__init__ = futures_data_init

# Patch custom venue
class BinanceLiveDataClientFactory(LiveDataClientFactory):
    @staticmethod
    def create(
        loop,
        name,
        config,
        msgbus,
        cache,
        clock,
        logger,
    ):
        base_url_http_default: str = _get_http_base_url(
            config.account_type, config.testnet, config.us
        )
        base_url_ws_default: str = _get_ws_base_url(config.account_type, config.testnet, config.us)

        client = get_cached_binance_http_client(
            loop=loop,
            clock=clock,
            logger=logger,
            key=config.api_key,
            secret=config.api_secret,
            account_type=config.account_type,
            base_url=config.base_url_http or base_url_http_default,
            is_testnet=config.testnet,
            is_us=config.us,
        )

        venue = Venue(name.upper())

        if config.account_type.is_spot or config.account_type.is_margin:
            # Get instrument provider singleton
            provider = get_cached_binance_spot_instrument_provider(
                client=client,
                logger=logger,
                venue=venue,
                account_type=config.account_type,
                config=config.instrument_provider,
            )

            # Create client
            return spot_data.BinanceSpotDataClient(
                loop=loop,
                client=client,
                msgbus=msgbus,
                cache=cache,
                clock=clock,
                logger=logger,
                instrument_provider=provider,
                venue=venue,
                account_type=config.account_type,
                base_url_ws=config.base_url_ws or base_url_ws_default,
            )
        else:
            # Get instrument provider singleton
            provider = get_cached_binance_futures_instrument_provider(
                client=client,
                logger=logger,
                venue=venue,
                account_type=config.account_type,
                config=config.instrument_provider,
            )

            # Create client
            return futures_data.BinanceFuturesDataClient(
                loop=loop,
                client=client,
                msgbus=msgbus,
                cache=cache,
                clock=clock,
                logger=logger,
                instrument_provider=provider,
                venue=venue,
                account_type=config.account_type,
                base_url_ws=config.base_url_ws or base_url_ws_default,
            )


# Patch custom account id and venue
class BinanceLiveExecClientFactory(LiveExecClientFactory):
    @staticmethod
    def create(
        loop,
        name,
        config,
        msgbus,
        cache,
        clock,
        logger,
    ):
        base_url_http_default: str = _get_http_base_url(
            config.account_type, config.testnet, config.us
        )
        base_url_ws_default: str = _get_ws_base_url(config.account_type, config.testnet, config.us)

        client = get_cached_binance_http_client(
            loop=loop,
            clock=clock,
            logger=logger,
            key=config.api_key,
            secret=config.api_secret,
            account_type=config.account_type,
            base_url=config.base_url_http or base_url_http_default,
            is_testnet=config.testnet,
        )

        res = name.partition("-")
        if res[1]:
            account_id = AccountId(name)
        else:
            account_id = AccountId(name, "-DEFAULT")
        venue = Venue(account_id.get_issuer())

        if config.account_type.is_spot or config.account_type.is_margin:
            # Get instrument provider singleton
            provider = get_cached_binance_spot_instrument_provider(
                client=client,
                logger=logger,
                venue=venue,
                account_type=config.account_type,
                config=config.instrument_provider,
            )

            # Create client
            return spot_execution.BinanceSpotExecutionClient(
                loop=loop,
                client=client,
                msgbus=msgbus,
                cache=cache,
                clock=clock,
                logger=logger,
                instrument_provider=provider,
                name=name,
                account_id=account_id,
                venue=venue,
                account_type=config.account_type,
                base_url_ws=config.base_url_ws or base_url_ws_default,
            )
        else:
            # Get instrument provider singleton
            provider = get_cached_binance_futures_instrument_provider(
                client=client,
                logger=logger,
                venue=venue,
                account_type=config.account_type,
                config=config.instrument_provider,
            )

            # Create client
            return futures_execution.BinanceFuturesExecutionClient(
                loop=loop,
                client=client,
                msgbus=msgbus,
                cache=cache,
                clock=clock,
                logger=logger,
                instrument_provider=provider,
                name=name,
                account_id=account_id,
                venue=venue,
                oms_type=config.oms_type,
                account_type=config.account_type,
                base_url_ws=config.base_url_ws or base_url_ws_default,
            )


def spot_provider_init(
    self,
    client,
    logger,
    venue,
    account_type: BinanceAccountType = BinanceAccountType.SPOT,
    config=None,
):
    super(spot_providers.BinanceSpotInstrumentProvider, self).__init__(
        venue=venue,
        logger=logger,
        config=config,
    )

    self._client = client
    self._account_type = account_type

    self._http_wallet = BinanceSpotWalletHttpAPI(self._client)
    self._http_market = BinanceSpotMarketHttpAPI(self._client)


async def load_all_async(self, filters: Optional[Dict] = None) -> None:
    """
    Load the latest instruments into the provider asynchronously, optionally
    applying the given filters.

    Parameters
    ----------
    filters : Dict, optional
        The venue specific instrument loading filters to apply.

    """
    filters_str = "..." if not filters else f" with filters {filters}..."
    self._log.info(f"Loading all instruments{filters_str}")

    # Get current commission rates
    try:
        fee_res = await self._http_wallet.trade_fees()
        fees = {s.symbol: s for s in fee_res}
    except BinanceClientError:
        self._log.error(
            "Cannot load instruments: API key authentication failed "
            "(this is needed to fetch the applicable account fee tier).",
        )
        fees = {}

    # Get exchange info for all assets
    exchange_info = await self._http_market.exchange_info()
    for symbol_info in exchange_info.symbols:
        self._parse_instrument(
            symbol_info=symbol_info,
            fees=fees.get(symbol_info.symbol),
            ts_event=millis_to_nanos(exchange_info.serverTime),
        )


async def load_ids_async(
    self,
    instrument_ids: List[InstrumentId],
    filters: Optional[Dict] = None,
) -> None:
    """
    Load the instruments for the given IDs into the provider, optionally
    applying the given filters.

    Parameters
    ----------
    instrument_ids: List[InstrumentId]
        The instrument IDs to load.
    filters : Dict, optional
        The venue specific instrument loading filters to apply.

    Raises
    ------
    ValueError
        If any `instrument_id.venue` is not equal to `self.venue`.

    """
    if not instrument_ids:
        self._log.info("No instrument IDs given for loading.")
        return

    # Check all instrument IDs
    for instrument_id in instrument_ids:
        PyCondition.equal(instrument_id.venue, self.venue, "instrument_id.venue", "self.venue")

    filters_str = "..." if not filters else f" with filters {filters}..."
    self._log.info(f"Loading instruments {instrument_ids}{filters_str}.")

    # Get current commission rates
    try:
        fee_res = await self._http_wallet.trade_fees()
        fees = {s.symbol: s for s in fee_res}
    except BinanceClientError:
        self._log.error(
            "Cannot load instruments: API key authentication failed "
            "(this is needed to fetch the applicable account fee tier).",
        )
        fees = {}

    # Extract all symbol strings
    symbols: List[str] = [instrument_id.symbol.value for instrument_id in instrument_ids]

    # Get exchange info for all assets
    exchange_info = await self._http_market.exchange_info(symbols=symbols)
    for symbol_info in exchange_info.symbols:
        self._parse_instrument(
            symbol_info=symbol_info,
            fees=fees.get(symbol_info.symbol),
            ts_event=millis_to_nanos(exchange_info.serverTime),
        )


async def load_async(self, instrument_id: InstrumentId, filters: Optional[Dict] = None):
    """
    Load the instrument for the given ID into the provider asynchronously, optionally
    applying the given filters.

    Parameters
    ----------
    instrument_id: InstrumentId
        The instrument ID to load.
    filters : Dict, optional
        The venue specific instrument loading filters to apply.

    Raises
    ------
    ValueError
        If `instrument_id.venue` is not equal to `self.venue`.

    """
    PyCondition.not_none(instrument_id, "instrument_id")
    PyCondition.equal(instrument_id.venue, self.venue, "instrument_id.venue", "self.venue")

    filters_str = "..." if not filters else f" with filters {filters}..."
    self._log.debug(f"Loading instrument {instrument_id}{filters_str}.")

    symbol = instrument_id.symbol.value

    # Get current commission rates
    try:
        fees = await self._http_wallet.trade_fee(symbol=instrument_id.symbol.value)
    except BinanceClientError:
        self._log.error(
            "Cannot load instruments: API key authentication failed "
            "(this is needed to fetch the applicable account fee tier).",
        )
        fees = {}

    # Get exchange info for asset
    exchange_info = await self._http_market.exchange_info(symbol=symbol)
    for symbol_info in exchange_info.symbols:
        self._parse_instrument(
            symbol_info=symbol_info,
            fees=fees,
            ts_event=millis_to_nanos(exchange_info.serverTime),
        )


spot_providers.BinanceSpotInstrumentProvider.__init__ = spot_provider_init
spot_providers.BinanceSpotInstrumentProvider.load_all_async = load_all_async
spot_providers.BinanceSpotInstrumentProvider.load_ids_async = load_ids_async
spot_providers.BinanceSpotInstrumentProvider.load_async = load_async


def futures_provider_init(
    self,
    client,
    logger,
    venue,
    account_type: BinanceAccountType = BinanceAccountType.FUTURES_USDT,
    config=None,
):
    super(futures_providers.BinanceFuturesInstrumentProvider, self).__init__(
        venue=venue,
        logger=logger,
        config=config,
    )

    self._client = client
    self._account_type = account_type

    self._http_wallet = BinanceFuturesWalletHttpAPI(self._client)
    self._http_market = BinanceFuturesMarketHttpAPI(self._client, account_type=account_type)


futures_providers.BinanceFuturesInstrumentProvider.__init__ = futures_provider_init



import asyncio
import re

from aiokafka import AIOKafkaProducer

from nacre.actors.pubsub cimport PubSub

from nacre.config import KafkaConfig

from nautilus_trader.common.logging cimport Logger
from nautilus_trader.core.correctness cimport Condition

# from nautilus_trader.model.events.order cimport OrderModifyRejected
# from nautilus_trader.model.events.order cimport OrderPendingCancel
# from nautilus_trader.model.events.order cimport OrderPendingUpdate
# from nautilus_trader.model.events.order cimport OrderRejected
# from nautilus_trader.model.events.order cimport OrderAccepted
# from nautilus_trader.model.events.order cimport OrderCanceled
# from nautilus_trader.model.events.order cimport OrderCancelRejected
# from nautilus_trader.model.events.order cimport OrderDenied
# from nautilus_trader.model.events.order cimport OrderExpired
# from nautilus_trader.model.events.order cimport OrderFilled
from nautilus_trader.model.events.order cimport OrderEvent
from nautilus_trader.model.events.order cimport OrderInitialized
from nautilus_trader.model.events.order cimport OrderSubmitted
from nautilus_trader.model.events.position cimport PositionEvent
from nautilus_trader.model.identifiers cimport TraderId
from nautilus_trader.model.orders.base cimport Order
from nautilus_trader.model.position cimport Position
from nautilus_trader.serialization.msgpack.serializer cimport MsgPackSerializer

from nacre.serialization.avro.serializer cimport AvroSerializer


# from nautilus_trader.model.events.order cimport OrderTriggered
# from nautilus_trader.model.events.order cimport OrderUpdated
# from nautilus_trader.model.events.position cimport PositionChanged
# from nautilus_trader.model.events.position cimport PositionClosed
# from nautilus_trader.model.events.position cimport PositionOpened

cdef class KafkaPubSub(PubSub):

    def __init__(self, config: KafkaConfig):
        Condition.type(config, KafkaConfig, "config")
        super().__init__(config)

        self._key = b""
        self.config = config

        # init kafka producer
        self._producer = None
        self._consumer = None

        self._serializer = AvroSerializer(config.schema_registry)

        self._start_producer_task = None
        self._stop_producer_task = None


    cpdef object wrap_message(self, object obj):
        if isinstance(obj, OrderEvent):
            if isinstance(obj, (OrderInitialized, OrderSubmitted)):
                return obj

            order = self.cache.order(obj.client_order_id)
            if order is not None:
                return (obj, order)
            else:
                return obj
        elif isinstance(obj, PositionEvent):
            position = self.cache.position(obj.position_id)
            if position is not None:
                return (obj, position)
            else:
                return obj
        else:
            return obj

    cpdef void relay(self, object obj) except *:
        self._loop.create_task(self._relay(obj))

    cpdef str detect_topic(self, object obj):
        if isinstance(obj, OrderEvent):
            return "events.order"
        elif isinstance(obj, PositionEvent):
            return "events.position"
        elif isinstance(obj, Order):
            return "states.order"
        elif isinstance(obj, Position):
            return "states.position"
        else:
            return "general"

    async def _relay(self, object obj):
        if isinstance(obj, tuple):
            for o in obj:
                await self._send(o)
        else:
            await self._send(obj)

    async def _send(self, object obj):
        topic = self.detect_topic(obj)
        cdef bytes obj_value = b""
        try:
            obj_value = await self._serializer.serialize(obj)
        except Exception as ex:
            self._log.exception("Error on sending object", ex)
            return

        await self._producer.send_and_wait(topic, value=obj_value, key=self._key)

    def connect(self):
        self._start_producer_task = self._loop.create_task(self._run_producer())
        self._key = self.trader_id.value.encode()

    def disconnect(self):
        self._stop_producer_task = self._loop.create_task(self._stop_producer())

    async def _run_producer(self):
        self._producer = AIOKafkaProducer(
            acks=1,
            loop=self._loop,
            bootstrap_servers=self.config.bootstrap_servers,
            security_protocol=self.config.security_protocol,
            client_id=self.trader_id.value,
        )

        await self._producer.start()
        self._log.info("Kafka pubsub started")

    async def _stop_producer(self):
        await self._producer.stop()
        await self._serializer.stop()
        self._log.info("Kafka pubsub stopped")



cdef class AsyncSerializer:
    async def serialize(self, object obj) -> bytes:
        pass

    async def deserialize(self, bytes obj_bytes) -> object:
        pass

    async def stop(self):
        pass



import asyncio
from datetime import timedelta
from typing import Awaitable, Callable, Dict, List  # noqa: TYP001

from aiohttp import ClientConnectorError
from nautilus_trader.common.clock import LiveClock
from nautilus_trader.common.logging import Logger
from nautilus_trader.common.timer import TimeEvent
from nautilus_trader.core.uuid import UUID4
from tenacity import retry
from tenacity.retry import retry_if_exception_type

from nacre.network.websocket import WebSocketClient


class ZbWebSocketClient(WebSocketClient):
    """
    Provides a `Zb` streaming WebSocket client.
    """

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        clock: LiveClock,
        logger: Logger,
        handler: Callable[[bytes], None],
        base_url: str,
    ):
        super().__init__(
            loop=loop,
            logger=logger,
            handler=handler,
            max_retry_connection=5,
        )

        self._base_url = base_url
        self._clock = clock
        self._subscriptions: Dict[str, Dict] = {}
        self._pager_name = None

        self._post_connect_callbacks: List[Awaitable] = []

    def add_post_connect_callbacks(self, *callbacks: Awaitable):
        self._post_connect_callbacks.extend(callbacks)

    async def post_connection(self):
        # Multiple writer should exist in other tasks
        # Ref: https://docs.aiohttp.org/en/stable/client_quickstart.html#websockets
        self._loop.create_task(self._post_connection())

    async def post_reconnection(self):
        self._loop.create_task(self._post_connection())

    async def _post_connection(self):
        if self._pager_name is None:
            self._set_up_pager()

        await self.on_post_connect()

    def _set_up_pager(self):
        self._pager_name = UUID4().value
        self._clock.set_timer(
            name=self._pager_name,
            interval=timedelta(seconds=3),  # Ping interval Hardcoded for now
            start_time=None,
            stop_time=None,
            callback=self._on_pager_inteval,
        )

    def _on_pager_inteval(self, event: TimeEvent):
        self._loop.create_task(self.ping())

    async def ping(self):
        if not self.retrying:
            try:
                await self.send_json({"action": "ping"})
            except ConnectionResetError:
                return

    async def on_post_connect(self):
        # Resubscribe when reconnect
        if self._subscriptions:
            await self._resubscribe()

        for callback_routine in self._post_connect_callbacks:
            await callback_routine

    async def post_disconnection(self):
        if self._pager_name:
            self._clock.cancel_timer(self._pager_name)
            self._pager_name = None

    async def _resubscribe(self):
        for subscription in self._subscriptions.values():
            self._log.debug(f"Resubscribe {subscription}")
            await self.send_json(subscription)

    @retry(retry=retry_if_exception_type(ClientConnectorError))
    async def connect(self, start: bool = True, **ws_kwargs) -> None:
        if "ws_url" in ws_kwargs:
            ws_kwargs.pop("ws_url")
        try:
            await super().connect(ws_url=self._base_url, start=start, **ws_kwargs)
        except ClientConnectorError as ex:
            self._log.warning(f"{ex}, Retrying...")
            raise ex

    # One time request
    async def _request_channel(self, channel: str, **kwargs):
        payload = {"channel": channel}
        if kwargs:
            payload.update(kwargs)
        await self._send_ws_request(payload)

    # Subscription continues to push when state change
    async def _subscribe_channel(self, channel: str, **kwargs):
        payload = {"channel": channel}
        if kwargs:
            payload.update(kwargs)

        # Resubscribe when reconnect
        hash_key = str(payload.values())
        self._subscriptions[hash_key] = payload
        await self._send_ws_request(payload)

    # Not tested
    async def _unsubscribe_channel(self, channel: str, **kwargs):
        payload = {"channel": channel}
        if kwargs:
            payload.update(kwargs)

        self._subscriptions.pop(channel, None)
        await self._send_ws_request(payload)

    async def _send_ws_request(self, payload: dict):
        while not self.is_connected:
            await self._sleep0()

        await self.send_json(payload)



# -------------------------------------------------------------------------------------------------
#  Copyright (C) 2015-2021 Nautech Systems Pty Ltd. All rights reserved.
#  https://nautechsystems.io
#
#  Licensed under the GNU Lesser General Public License Version 3.0 (the "License");
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at https://www.gnu.org/licenses/lgpl-3.0.en.html
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# -------------------------------------------------------------------------------------------------

import asyncio

from libc.stdint cimport int64_t
from nautilus_trader.core.uuid cimport UUID4
from nautilus_trader.common.logging cimport LogColor
from nautilus_trader.core.correctness cimport Condition
from nautilus_trader.execution.reports cimport ExecutionMassStatus
from nautilus_trader.execution.reports cimport OrderStatusReport
from nautilus_trader.execution.reports cimport TradeReport
from nautilus_trader.live.execution_client cimport LiveExecutionClient as NautilusLiveExecutionClient
from nautilus_trader.model.identifiers cimport AccountId
from nautilus_trader.model.identifiers cimport InstrumentId


cdef class LiveExecutionClient(NautilusLiveExecutionClient):
    cpdef void _set_venue(self, Venue venue) except *:
        Condition.not_none(venue, "venue")
        self.venue = venue

    cpdef void _set_account_id(self, AccountId account_id) except *:
        Condition.not_none(account_id, "account_id")
        # Override check
        # Condition.equal(self.id.value, account_id.issuer, "id.value", "account_id.issuer")

        self.account_id = account_id

    async def generate_mass_status(self, lookback_mins: Optional[int]):
        full_mass_status = await super().generate_mass_status(lookback_mins)

        cdef ExecutionMassStatus mass_status = ExecutionMassStatus(
            client_id=self.id,
            account_id=self.account_id,
            venue=self.venue,
            report_id=UUID4(),
            ts_init=self._clock.timestamp_ns(),
        )

        instrument_id_filters = self._instrument_provider._load_ids_on_start
        if instrument_id_filters is None:
            return full_mass_status

        instrument_ids = [InstrumentId.from_str_c(i) for i in instrument_id_filters]

        self._log.info(f"Filtering ExecutionMassStatus by instrument ids {instrument_id_filters}.")
        # Filter mass_status with instrument only specified on providers
        order_reports = []
        for order_report in full_mass_status.order_reports().values():
            if order_report.instrument_id in instrument_ids:
                order_reports.append(order_report)

        trade_reports = []
        for trade_reports_group in full_mass_status.trade_reports().values():
            if trade_reports_group[0].instrument_id in instrument_ids:
                trade_reports.extend(trade_reports_group)

        position_reports = []
        for instrument_id, position_reports_group in full_mass_status.position_reports().items():
            if instrument_id in instrument_ids:
                position_reports.extend(position_reports_group)

        mass_status.add_order_reports(reports=order_reports)
        mass_status.add_trade_reports(reports=trade_reports)
        mass_status.add_position_reports(reports=position_reports)

        return mass_status

    cpdef void _send_order_status_report(self, OrderStatusReport report) except *:
        # Filter report first
        instrument_id_filters = self._instrument_provider._load_ids_on_start
        if instrument_id_filters is not None:
            instrument_ids = [InstrumentId.from_str_c(i) for i in instrument_id_filters]
            if report.instrument_id not in instrument_ids:
                self._log.info(f"Ignoring {report}.", color=LogColor.BLUE)
                return

        self._msgbus.send(
            endpoint="ExecEngine.reconcile_report",
            msg=report,
        )

    cpdef void _send_trade_report(self, TradeReport report) except *:
        # Filter report first
        instrument_id_filters = self._instrument_provider._load_ids_on_start
        if instrument_id_filters is not None:
            instrument_ids = [InstrumentId.from_str_c(i) for i in instrument_id_filters]
            if report.instrument_id not in instrument_ids:
                self._log.info(f"Ignoring {report}.", color=LogColor.BLUE)
                return

        self._msgbus.send(
            endpoint="ExecEngine.reconcile_report",
            msg=report,
        )




from nautilus_trader.common.actor cimport Actor
from nautilus_trader.common.logging cimport LoggerAdapter
from nautilus_trader.common.queue cimport Queue
from nautilus_trader.model.identifiers cimport TraderId


cdef class PubSub(Actor):
    cdef list topic_filters
    cdef object _loop
    cdef object _run_queues_task
    cdef Queue _queue

    cdef readonly bint is_running

    cpdef int qsize(self) except *

    cpdef void kill(self) except *
    cdef void _enqueue_sentinels(self) except *

    cpdef object wrap_message(self, object obj)
    cpdef void relay(self, object obj) except *



import asyncio
from typing import Any, Dict, List, Optional

from nautilus_trader.core.correctness import PyCondition

from nacre.adapters.zb.common import format_endpoint
from nacre.adapters.zb.common import format_symbol
from nacre.adapters.zb.http.client import ZbHttpClient


class ZbFutureAccountHttpAPI:

    BASE_ENDPOINT = "/{quote}/Server/api/v2/"
    FUTURE_ACCOUNT_TYPES = (1, 2)

    def __init__(self, client: ZbHttpClient):
        """
        Initialize a new instance of the ``ZbFutureAccountHttpAPI`` class.

        Parameters
        ----------
        client : ZbHttpClient
            The Binance REST API client.

        """
        PyCondition.not_none(client, "client")

        self.client = client

    async def new_order(
        self,
        symbol: str,
        side: int,
        amount: float,
        price: Optional[float] = None,
        action: Optional[int] = None,
        client_order_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Submit a new order.

        Parameters
        ----------
        symbol : str
            The symbol for the request.
        side : int
            The order side
            1 open long (dual side position mode)
            2 open short (dual side position mode)
            3 close long (dual side position mode)
            4 close short (dual side position mode)
            5 buy (one side position mode)
            6 sell (one side position mode)
            0 reduce only
        amount : str
            The order quantity in base asset units for the request.
        price : str, optional
            The order price for the request, optional if action in [11, 12, 31, 52, 52]
        client_order_id : str, optional
            The client order ID for the request. A unique ID among open orders.
            Automatically generated if not provided.
        action : int, optional
            Order Type
            1 limit
            3 IOC
            4 maker only
            5 FOK

        Returns
        -------
        dict[str, Any]

        """
        payload: Dict[str, Any] = {"symbol": format_symbol(symbol), "side": side, "amount": amount}
        if price is not None:
            payload["price"] = price
        if action is not None:
            payload["action"] = action
        if client_order_id is not None:
            payload["clientOrderId"] = client_order_id

        return await self.client.sign_request(
            http_method="POST",
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "trade/order",
            payload=payload,
        )

    async def new_batch_order(  # noqa: C901
        self,
        orders: List[Dict[str, Any]],
    ) -> Dict[str, Any]:
        """
        Submit orders in batch

        Parameters
        ----------
        orders: List[Dict]
            Each order dict as follow
            symbol : str
                The symbol for the request.
            side : int
                The order side
                1 open long (dual side position mode)
                2 open short (dual side position mode)
                3 close long (dual side position mode)
                4 close short (dual side position mode)
                5 buy (one side position mode)
                6 sell (one side position mode)
                0 reduce only
            amount : str
                The order quantity in base asset units for the request.
            price : str, optional
                The order price for the request, optional if action in [11, 12, 31, 52, 52]
            orderCode: str
                The client order ID for the request. A unique ID among open orders.
                Automatically generated if not provided.
            action : int, optional
                Order Type
                1 limit
                3 IOC
                4 maker only
                5 FOK

        Returns
        -------
        dict[str, Any]

        """
        first_symbol = orders[0]["symbol"]
        for order in orders:
            order["symbol"] = format_symbol(order["symbol"])

        payload: Dict[str, Any] = {"orderDatas": orders}

        return await self.client.sign_request(
            http_method="POST",
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=first_symbol) + "trade/batchOrder",
            payload=payload,
        )

    async def cancel_order(
        self,
        symbol: str,
        order_id: Optional[str] = None,
        client_order_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Cancel specified orders on a symbol.

        Parameters
        ----------
        symbol : str
            The symbol for the request.
        order_id: str, optional
            OrderId return from exchange. optional if client_order_id set
        client_order_id: str, optional
            ClientOrderId generated by client. optional if order_id set

        Returns
        -------
        dict[str, Any]

        """
        payload: Dict[str, Any] = {"symbol": format_symbol(symbol)}
        if order_id is not None:
            payload["orderId"] = order_id
        elif client_order_id is not None:
            payload["clientOrderId"] = client_order_id

        return await self.client.sign_request(
            http_method="POST",
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "trade/cancelOrder",
            payload=payload,
        )

    async def cancel_batch_order(
        self,
        symbol: str,
        order_ids: Optional[List[str]] = None,
        client_order_ids: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """
        Cancel specified orders on a symbol.

        Parameters
        ----------
        symbol : str
            The symbol for the request.
        order_ids: List[str], optional
            OrderId list return from exchange. optional if client_order_id set
        client_order_ids: List[str], optional
            ClientOrderId list generated by client. optional if order_id set

        Returns
        -------
        dict[str, Any]

        """
        payload: Dict[str, Any] = {"symbol": format_symbol(symbol)}
        if order_ids is not None:
            payload["orderIds"] = order_ids
        elif client_order_ids is not None:
            payload["clientOrderIds"] = client_order_ids

        return await self.client.sign_request(
            http_method="POST",
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "trade/batchCancelOrder",
            payload=payload,
        )

    async def cancel_open_orders(self, symbol: str) -> Dict[str, Any]:
        """
        Cancel all active orders on a symbol.

        Parameters
        ----------
        symbol : str
            The symbol for the request.

        Returns
        -------
        dict[str, Any]

        """
        payload: Dict[str, Any] = {"symbol": format_symbol(symbol)}

        return await self.client.sign_request(
            http_method="POST",
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "trade/cancelAllOrders",
            payload=payload,
        )

    async def get_open_orders(
        self,
        symbol: str,
        page_num: Optional[int] = None,
        page_size: Optional[int] = None,
        type: Optional[int] = 0,
        side: Optional[int] = None,
        action: Optional[int] = None,
    ) -> Dict[str, Any]:
        """
        Get all orders on a symbol.


        Parameters
        ----------
        symbol : str
            The symbol for the request.
        type : int, optional
            -1 sell; 1 buy, 0 all
        side : int, optional
            The order side
            1 open long (dual side position mode)
            2 open short (dual side position mode)
            3 close long (dual side position mode)
            4 close short (dual side position mode)
            5 buy (one side position mode)
            6 sell (one side position mode)
            0 reduce only
        action : int, optional
            Order Type
            1 limit
            3 IOC
            4 maker only
            5 FOK
        page_num : int, optional
            page number
        page_size : int, optional
            page size

        Returns
        -------
        dict[str, Any]

        """
        payload: Dict[str, Any] = {"symbol": format_symbol(symbol)}
        if type is not None:
            payload["type"] = type
        if side is not None:
            payload["side"] = side
        if action is not None:
            payload["action"] = action
        if page_num is not None:
            payload["pageNum"] = page_num
        if page_size is not None:
            payload["pageSize"] = page_size

        return await self.client.sign_request(
            http_method="GET",
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "trade/getUndoneOrders",
            payload=payload,
        )

    async def get_all_orders(
        self,
        symbol: str,
        type: Optional[int] = None,
        side: Optional[int] = None,
        date_range: Optional[int] = None,
        action: Optional[int] = None,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        page_num: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        """
        Get all orders on a symbol.


        Parameters
        ----------
        symbol : str
            The symbol for the request.
        type : int, optional
            -1 sell; 1 buy, 0 all
        side : int, optional
            The order side
            1 open long (dual side position mode)
            2 open short (dual side position mode)
            3 close long (dual side position mode)
            4 close short (dual side position mode)
            5 buy (one side position mode)
            6 sell (one side position mode)
            0 reduce only
        date_range : int, optional
            0 recent; 1 more
        action : int, optional
            Order Type
            1 limit
            3 IOC
            4 maker only
            5 FOK
        start_time : int, optional
            Start time
        end_time : int, optional
            End time
        page_num : int, optional
            page number
        page_size : int, optional
            page size

        Returns
        -------
        dict[str, Any]

        """
        payload: Dict[str, Any] = {"symbol": format_symbol(symbol)}
        if type is not None:
            payload["type"] = type
        if side is not None:
            payload["side"] = side
        if action is not None:
            payload["action"] = action
        if date_range is not None:
            payload["dateRange"] = date_range
        if start_time is not None:
            payload["startTime"] = start_time
        if end_time is not None:
            payload["endTime"] = end_time
        if page_num is not None:
            payload["pageNum"] = page_num
        if page_size is not None:
            payload["pageSize"] = page_size

        return await self.client.sign_request(
            http_method="GET",
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "trade/getAllOrders",
            payload=payload,
        )

    async def get_order(
        self,
        symbol: str,
        order_id: Optional[str] = None,
        client_order_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Get specified orders.

        Parameters
        ----------
        symbol : str
            The symbol for the request.
        order_id: str, optional
            OrderId return from exchange. optional if client_order_id set
        client_order_id: str, optional
            ClientOrderId generated by client. optional if order_id set

        Returns
        -------
        dict[str, Any]

        """
        payload: Dict[str, Any] = {"symbol": format_symbol(symbol)}
        if order_id is not None:
            payload["orderId"] = order_id
        elif client_order_id is not None:
            payload["clientOrderId"] = client_order_id

        return await self.client.sign_request(
            http_method="GET",
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "trade/getOrder",
            payload=payload,
        )

    async def get_trades(
        self,
        symbol: str,
        order_id: str,
        page_num: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:

        payload: Dict[str, Any] = {"symbol": format_symbol(symbol), "orderId": order_id}
        if page_num is not None:
            payload["pageNum"] = page_num
        if page_size is not None:
            payload["pageSize"] = page_size

        return await self.client.sign_request(
            http_method="GET",
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "trade/getTradeList",
            payload=payload,
        )

    async def get_history_trades(
        self,
        symbol: str,
        side: Optional[int] = None,
        date_range: Optional[int] = None,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        page_num: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        """
        Get all orders on a symbol.


        Parameters
        ----------
        symbol : str
            The symbol for the request.
        type : int, optional
            -1 sell; 1 buy, 0 all
        side : int, optional
            OrderSide 0 all; 1 open long; 2 open short; 3 close long; 4 close short
        date_range : int, optional
            0 recent; 1 more
        action : int, optional
            Order Type
            1 limit
            3 IOC
            4 maker only
            5 FOK
        start_time : int, optional
            Start time
        end_time : int, optional
            End time
        page_num : int, optional
            page number
        page_size : int, optional
            page size

        Returns
        -------
        dict[str, Any]

        """
        payload: Dict[str, Any] = {"symbol": format_symbol(symbol)}
        if side is not None:
            payload["side"] = side
        if date_range is not None:
            payload["dateRange"] = date_range
        if start_time is not None:
            payload["startTime"] = start_time
        if end_time is not None:
            payload["endTime"] = end_time
        if page_num is not None:
            payload["pageNum"] = page_num
        if page_size is not None:
            payload["pageSize"] = page_size

        return await self.client.sign_request(
            http_method="GET",
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "trade/tradeHistory",
            payload=payload,
        )

    # TODO: add Algo order api

    async def get_bill(
        self,
        future_account_type: int = 1,
        currency_id: Optional[int] = None,
        currency_name: Optional[str] = None,
        type: Optional[int] = None,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        page_num: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {"futuresAccountType": future_account_type}
        if currency_id is not None:
            payload["currencyId"] = currency_id
        if currency_name is not None:
            payload["currencyName"] = currency_name
        if type is not None:
            payload["type"] = type
        if start_time is not None:
            payload["startTime"] = start_time
        if end_time is not None:
            payload["endTime"] = end_time
        if page_num is not None:
            payload["pageNum"] = page_num
        if page_size is not None:
            payload["pageSize"] = page_size
        return await self.client.sign_request(
            http_method="GET",
            url_path=format_endpoint(self.BASE_ENDPOINT, future_account_type=future_account_type)
            + "Fund/getBill",
            payload=payload,
        )

    async def get_account(
        self, future_account_type: Optional[int] = None, convert_unit: Optional[str] = None
    ) -> Dict[str, Any]:
        if future_account_type is None:
            reply: Dict[str, Any] = {"data": {"accounts": [], "assets": []}}
            aws = [self._get_account(future_account_type=ft) for ft in self.FUTURE_ACCOUNT_TYPES]
            results = await asyncio.gather(*aws)
            for ret in results:
                reply["data"]["accounts"].append(ret["data"]["account"])
                reply["data"]["assets"].extend(ret["data"]["assets"])
            return reply
        else:
            reply = await self._get_account(future_account_type=future_account_type)
            reply["data"]["accounts"] = [reply["data"]["account"]]
            return reply

    async def _get_account(
        self, future_account_type: int, convert_unit: Optional[str] = None
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {"futuresAccountType": future_account_type}
        if convert_unit is not None:
            payload["convert_unit"] = convert_unit

        return await self.client.sign_request(
            http_method="GET",
            url_path=format_endpoint(self.BASE_ENDPOINT, future_account_type=future_account_type)
            + "Fund/getAccount",
            payload=payload,
        )

    async def get_positions(
        self,
        future_account_type: Optional[int] = None,
        symbol: Optional[str] = None,
        side: Optional[int] = None,
    ) -> Dict[str, Any]:
        if future_account_type is None:
            reply: Dict[str, Any] = {"data": []}
            aws = [self._get_positions(ft, symbol, side) for ft in self.FUTURE_ACCOUNT_TYPES]
            results = await asyncio.gather(*aws)
            for ret in results:
                reply["data"].extend(ret["data"])
            return reply
        else:
            reply = await self._get_positions(future_account_type, symbol, side)
            return reply

    async def _get_positions(
        self,
        future_account_type: int,
        symbol: Optional[str] = None,
        side: Optional[int] = None,
    ) -> Dict[str, Any]:
        """
        Get position info

        Parameters
        ----------
        symbol : str, optional
            The market symbol.
        side : int, optional
            1 for LONG, 0 for SHORT
        future_account_type: int, optional, default 1
            1 for USDT perpetual

        Returns
        -------
        dict[str, Any]

        """
        payload: Dict[str, Any] = {"futuresAccountType": future_account_type}
        if symbol is not None:
            payload["symbol"] = format_symbol(symbol)
        if side is not None:
            payload["side"] = side

        return await self.client.sign_request(
            http_method="GET",
            url_path=format_endpoint(self.BASE_ENDPOINT, future_account_type=future_account_type)
            + "Positions/getPositions",
            payload=payload,
        )

    async def margin_info(
        self,
        positions_id: int,
        future_account_type: int = 1,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "futuresAccountType": future_account_type,
            "positionsId": positions_id,
        }

        return await self.client.sign_request(
            http_method="GET",
            url_path=format_endpoint(self.BASE_ENDPOINT, future_account_type=future_account_type)
            + "Positions/marginInfo",
            payload=payload,
        )

    async def margin_history(
        self,
        symbol: Optional[str] = None,
        type: Optional[int] = None,
        future_account_type: int = 1,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "futuresAccountType": future_account_type,
        }
        if symbol is not None:
            payload["symbol"] = format_symbol(symbol)
        if type is not None:
            payload["type"] = type

        return await self.client.sign_request(
            http_method="GET",
            url_path=format_endpoint(self.BASE_ENDPOINT, future_account_type=future_account_type)
            + "Fund/marginHistory",
            payload=payload,
        )

    async def update_margin(
        self,
        positions_id: int,
        amount: float,
        type: int,
        future_account_type: int = 1,
    ) -> Dict[str, Any]:
        """
        Change margin on specified position

        Parameters
        ----------
        positions_id : int
            The position id.
        amount : float
            The margin amount.
        type : int
            1 for increase; 0 for decrease
        future_account_type: int, optional, default 1
            1 for USDT perpetual

        Returns
        -------
        dict[str, Any]

        """
        payload: Dict[str, Any] = {
            "futuresAccountType": future_account_type,
            "positionsId": positions_id,
            "type": type,
            "amount": amount,
        }

        return await self.client.sign_request(
            http_method="POST",
            url_path=format_endpoint(self.BASE_ENDPOINT, future_account_type=future_account_type)
            + "Positions/updateMargin",
            payload=payload,
        )

    async def update_leverage(
        self,
        symbol: str,
        leverage: int,
        future_account_type: int = 1,
    ) -> Dict[str, Any]:
        """
        Change leverage on specified market

        Parameters
        ----------
        symbol : str
            The market symbol.
        leverage : int
            The leverage.
        future_account_type: int, optional, default 1
            1 for USDT perpetual
            2 for QC perpetual

        Returns
        -------
        dict[str, Any]

        """
        payload: Dict[str, Any] = {
            "futuresAccountType": future_account_type,
            "symbol": format_symbol(symbol),
            "leverage": leverage,
        }

        return await self.client.sign_request(
            http_method="POST",
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "setting/setLeverage",
            payload=payload,
        )

    async def update_position_mode(
        self,
        symbol: str,
        mode: int,
        future_account_type: int = 1,
    ) -> Dict[str, Any]:
        """
        Change position mode on specified market

        Parameters
        ----------
        symbol : str
            The market symbol.
        mode : int
            1 for one side, 2 for dual side
        future_account_type: int, optional, default 1
            1 for USDT perpetual
            2 for QC perpetual
            3 for COIN-based perpetual

        Returns
        -------
        dict[str, Any]

        """

        payload: Dict[str, Any] = {
            "futuresAccountType": future_account_type,
            "symbol": symbol,
            "positionsMode": mode,
        }
        return await self.client.sign_request(
            http_method="POST",
            url_path=format_endpoint(self.BASE_ENDPOINT, future_account_type=future_account_type)
            + "setting/setPositionsMode",
            payload=payload,
        )

    async def get_nominal_value(
        self,
        symbol: str,
        side: Optional[int] = None,
        future_account_type: int = 1,
    ) -> Dict[str, Any]:
        """
        Change leverage on specified market

        Parameters
        ----------
        symbol : str
            The market symbol.
        side : int, optional
            1 for LONG, 0 for SHORT
        future_account_type: int, optional, default 1
            1 for USDT perpetual
            2 for QC perpetual

        Returns
        -------
        dict[str, Any]

        """
        payload: Dict[str, Any] = {
            "futuresAccountType": future_account_type,
            "symbol": format_symbol(symbol),
        }
        if side is not None:
            payload["side"] = side

        return await self.client.sign_request(
            http_method="GET",
            url_path=format_endpoint(self.BASE_ENDPOINT, future_account_type=future_account_type)
            + "Positions/getNominalValue",
            payload=payload,
        )

    async def set_auto_append_margin(
        self,
        positions_id: int,
        max_additional_usd: float,
        future_account_type: int = 1,
    ) -> Dict[str, Any]:
        """
        Enable/Disable auto append margin

        Parameters
        ----------
        positions_id : int
            The position id.
        max_additional_usd : float
            USD value to append, 0 for disable
        future_account_type: int, optional, default 1
            1 for USDT perpetual

        Returns
        -------
        dict[str, Any]

        """
        payload: Dict[str, Any] = {
            "futuresAccountType": future_account_type,
            "maxAdditionalUSDValue": max_additional_usd,
            "positionsId": positions_id,
        }

        return await self.client.sign_request(
            http_method="POST",
            url_path=format_endpoint(self.BASE_ENDPOINT, future_account_type=future_account_type)
            + "Positions/updateAppendUSDValue",
            payload=payload,
        )

    async def transfer_fund(
        self,
        currency_name: str,
        amount: float,
        client_id: str,
        side: int,
    ) -> Dict[str, Any]:
        """
        Transfer fund between spot and future

        Parameters
        ----------
        currency_name : str
            The asset name.
        amount : float
            Transfer amount.
        client_id: str
            Unique client id, max len 18
        side: int
            1 for deposit, 0 for withdrawal

        Returns
        -------
        dict[str, Any]

        """
        payload: Dict[str, Any] = {
            "currencyName": currency_name,
            "amount": amount,
            "clientId": client_id,
            "side": side,
        }

        return await self.client.sign_request(
            http_method="POST",
            url_path=format_endpoint(self.BASE_ENDPOINT, currency=currency_name)
            + "Fund/transferFund",
            payload=payload,
        )



import asyncio
from typing import Any, Dict, List, Optional

from nautilus_trader.core.correctness import PyCondition

from nacre.adapters.zb.common import format_endpoint
from nacre.adapters.zb.common import format_symbol
from nacre.adapters.zb.http.client import ZbHttpClient


class ZbFutureMarketHttpAPI:
    """
    Provides access to the `ZB FUTURE Market` HTTP REST API.
    """

    BASE_ENDPOINT = "/{quote}/api/public/v1/"

    def __init__(self, client: ZbHttpClient):
        """
        Initialize a new instance of the ``ZbFutureMarketHttpAPI`` class.

        Parameters
        ----------
        client : ZbHttpClient
            The Binance REST API client.

        """
        PyCondition.not_none(client, "client")

        self.client = client

    async def market_list(self) -> List[Dict]:
        results = await asyncio.gather(
            self.client.query(url_path="/Server/api/v2/" + "config/marketList"),
            self.client.query(url_path="/qc/Server/api/v2/" + "config/marketList"),
        )
        markets = []
        for res in results:
            markets.extend(res.get("data", []))
        return markets

    async def depth(
        self, symbol: str, size: Optional[int] = None, scale: Optional[int] = None
    ) -> Dict[str, Any]:
        payload: Dict[str, str] = {"symbol": format_symbol(symbol)}
        if size is not None:
            payload["size"] = str(size)
        if scale is not None:
            payload["scale"] = str(scale)

        return await self.client.query(
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "depth",
            payload=payload,
        )

    async def trades(self, symbol: str, size: Optional[int] = None) -> Dict[str, Any]:
        payload: Dict[str, str] = {"symbol": format_symbol(symbol)}
        if size is not None:
            payload["size"] = str(size)

        return await self.client.query(
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "trade",
            payload=payload,
        )

    async def klines(
        self,
        symbol: str,
        period: str,
        size: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, str] = {
            "symbol": format_symbol(symbol),
            "period": period,
        }
        if size is not None:
            payload["size"] = str(size)

        return await self.client.query(
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "kline",
            payload=payload,
        )

    async def mark_klines(
        self,
        symbol: str,
        period: str,
        size: Optional[int] = None,
    ) -> List[List[Any]]:
        payload: Dict[str, str] = {
            "symbol": format_symbol(symbol),
            "period": period,
        }
        if size is not None:
            payload["size"] = str(size)

        return await self.client.query(
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "markKline",
            payload=payload,
        )

    async def index_klines(
        self,
        symbol: str,
        period: str,
        size: Optional[int] = None,
    ) -> List[List[Any]]:
        payload: Dict[str, str] = {
            "symbol": format_symbol(symbol),
            "period": period,
        }
        if size is not None:
            payload["size"] = str(size)

        return await self.client.query(
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "indexKline",
            payload=payload,
        )

    async def ticker_price(self, symbol: str = None) -> Dict[str, Any]:
        payload: Dict[str, str] = {}
        if symbol is not None:
            payload["symbol"] = format_symbol(symbol)

        return await self.client.query(
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "ticker",
            payload=payload,
        )

    async def mark_price(self, symbol: str = None) -> Dict[str, Any]:
        payload: Dict[str, str] = {}
        if symbol is not None:
            payload["symbol"] = format_symbol(symbol)

        return await self.client.query(
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "markPrice",
            payload=payload,
        )

    async def index_price(self, symbol: str = None) -> Dict[str, Any]:
        payload: Dict[str, str] = {}
        if symbol is not None:
            payload["symbol"] = format_symbol(symbol)

        return await self.client.query(
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "indexPrice",
            payload=payload,
        )

    async def premium_index(self, symbol: str = None) -> Dict[str, Any]:
        payload: Dict[str, str] = {}
        if symbol is not None:
            payload["symbol"] = format_symbol(symbol)

        return await self.client.query(
            url_path=format_endpoint("/{quote}/Server/api/v2/", symbol=symbol) + "premiumIndex",
            payload=payload,
        )

    async def historical_funding_rate(
        self,
        symbol: str = None,
        start_time_ms: Optional[int] = None,
        end_time_ms: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, str] = {}
        if symbol is not None:
            payload["symbol"] = format_symbol(symbol)
        if start_time_ms is not None:
            payload["startTime"] = str(start_time_ms)
        if end_time_ms is not None:
            payload["endTime"] = str(end_time_ms)
        if limit is not None:
            payload["limit"] = str(limit)

        return await self.client.query(
            url_path=format_endpoint("/{quote}/Server/api/v2/", symbol=symbol) + "fundingRate",
            payload=payload,
        )

    async def historical_liquidated_orders(
        self,
        symbol: str = None,
        start_time_ms: Optional[int] = None,
        end_time_ms: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, str] = {}
        if symbol is not None:
            payload["symbol"] = format_symbol(symbol)
        if start_time_ms is not None:
            payload["startTime"] = str(start_time_ms)
        if end_time_ms is not None:
            payload["endTime"] = str(end_time_ms)
        if limit is not None:
            payload["limit"] = str(limit)

        return await self.client.query(
            url_path=format_endpoint("/{quote}/Server/api/v2/", symbol=symbol) + "allForceOrders",
            payload=payload,
        )

    async def funding_rate(self, symbol: str) -> Dict[str, Any]:
        payload: Dict[str, str] = {"symbol": format_symbol(symbol)}

        return await self.client.query(
            url_path=format_endpoint(self.BASE_ENDPOINT, symbol=symbol) + "fundingRate",
            payload=payload,
        )



from nautilus_trader.common.actor cimport Actor
from nautilus_trader.common.component cimport Component
from nautilus_trader.common.logging cimport LogColor
from nautilus_trader.common.logging cimport Logger
from nautilus_trader.common.logging cimport LoggerAdapter
from nautilus_trader.core.correctness cimport Condition
from nautilus_trader.model.data.base cimport GenericData
from nautilus_trader.model.data.tick cimport QuoteTick
from nautilus_trader.model.data.tick cimport TradeTick
from nautilus_trader.trading.trader cimport Trader

from nacre.model.data.tick cimport MarkTick


cdef class AccessLoggerAdapter(LoggerAdapter):
    cpdef void info(self, str msg, LogColor color=*, dict extra=*) except *

cdef class Exposer(Actor):
    cdef object _loop
    cdef object _runner
    cdef readonly Trader trader
    """The trader back reference.\n\n:returns: `Trader`"""

    cdef object _run_http_server_task

    cpdef void register_trader(self, Trader trader) except *



from enum import Enum
from enum import unique


@unique
class ZbAccountType(Enum):
    """Represents a `Zb` account type"""

    SPOT = "SPOT"
    MARGIN = "MARGIN"
    FUTURES_USDT = "FUTURES_USDT"
    FUTURES_COIN = "FUTURES_COIN"

    @property
    def is_spot(self):
        return self == ZbAccountType.SPOT

    @property
    def is_margin(self):
        return self == ZbAccountType.MARGIN

    @property
    def is_futures(self) -> bool:
        return self in (ZbAccountType.FUTURES_USDT, ZbAccountType.FUTURES_COIN)



from decimal import Decimal

from nautilus_trader.adapters.binance.futures.parsing import execution
from nautilus_trader.adapters.binance.futures.schemas.account import BinanceFuturesPositionRisk
from nautilus_trader.core.uuid import UUID4
from nautilus_trader.model.enums import PositionSide
from nautilus_trader.model.identifiers import AccountId
from nautilus_trader.model.identifiers import InstrumentId
from nautilus_trader.model.objects import Price
from nautilus_trader.model.objects import Quantity

from nacre.execution.reports import PositionStatusReport


# Patch default binance futures execution position reports, added entry price and last price


def parse_position_report_http(
    account_id: AccountId,
    instrument_id: InstrumentId,
    data: BinanceFuturesPositionRisk,
    report_id: UUID4,
    ts_init: int,
) -> PositionStatusReport:
    net_size = Decimal(data.positionAmt)

    if net_size > 0:
        position_side = PositionSide.LONG
    elif net_size < 0:
        position_side = PositionSide.SHORT
    else:
        position_side = PositionSide.FLAT

    return PositionStatusReport(
        account_id=account_id,
        instrument_id=instrument_id,
        position_side=position_side,
        quantity=Quantity.from_str(str(abs(net_size))),
        report_id=report_id,
        entry_px=Price.from_str("%.4f" % float(data.entryPrice)),  # noqa: S001
        last_px=Price.from_str(data.markPrice),
        ts_last=ts_init,
        ts_init=ts_init,
    )


execution.parse_position_report_http = parse_position_report_http



from cpython.datetime cimport datetime
from nautilus_trader.indicators.base.indicator cimport Indicator


cdef class Avl(Indicator):
    cdef object _value_inputs
    cdef object _volume_inputs

    cdef readonly int period
    """The moving average period.\n\n:returns: `PriceType`"""
    cdef readonly int count
    """The count of inputs received by the indicator.\n\n:returns: `int`"""
    cdef readonly double value
    """The current value.\n\n:returns: `double`"""

    cpdef void update_raw(self, double value, double volume) except *

    cpdef void _increment_count(self) except *




import asyncio
from decimal import Decimal

from nautilus_trader.core.uuid cimport UUID4
from nautilus_trader.cache.cache cimport Cache
from nautilus_trader.common.clock cimport LiveClock
from nautilus_trader.common.logging cimport Logger
from nautilus_trader.model.position cimport Position
from nautilus_trader.msgbus.bus cimport MessageBus

from typing import Optional

from nacre.config.live import LiveExecEngineConfig

from nautilus_trader.execution.reports cimport OrderStatusReport
from nautilus_trader.execution.reports cimport PositionStatusReport
from nautilus_trader.execution.reports cimport TradeReport
from nautilus_trader.live.execution_engine cimport LiveExecutionEngine as NautilusLiveExecutionEngine
from nautilus_trader.model.c_enums.liquidity_side cimport LiquiditySide
from nautilus_trader.model.c_enums.order_side cimport OrderSide
from nautilus_trader.model.c_enums.order_status cimport OrderStatus
from nautilus_trader.model.c_enums.order_type cimport OrderType
from nautilus_trader.model.c_enums.position_side cimport PositionSide
from nautilus_trader.model.c_enums.time_in_force cimport TimeInForce
from nautilus_trader.model.currency cimport Currency
from nautilus_trader.model.identifiers cimport TradeId
from nautilus_trader.model.identifiers cimport VenueOrderId
from nautilus_trader.model.objects cimport Money
from nautilus_trader.model.objects cimport Price


# from nacre.execution.reports cimport PositionStatusReport

cdef class LiveExecutionEngine(NautilusLiveExecutionEngine):
    def __init__(
        self,
        loop not None: asyncio.AbstractEventLoop,
        MessageBus msgbus not None,
        Cache cache not None,
        LiveClock clock not None,
        Logger logger not None,
        config: Optional[LiveExecEngineConfig]=None,
    ):
        super().__init__(loop, msgbus, cache, clock, logger, config)
        self.reconcile_by_fake_fills = config.reconcile_by_fake_fills if config else False

    cdef bint _reconcile_position_report_hedging(self, PositionStatusReport report) except *:
        if self.reconcile_by_fake_fills:
            self._generate_fake_fills(report)

        cdef Position position = self._cache.position(report.venue_position_id)
        if position is None:
            self._log.error(
                f"Cannot reconcile position: "
                f"position ID {report.venue_position_id} not found.",
            )
            return False  # Failed
        if position.net_qty != report.net_qty:
            self._log.error(
                f"Cannot reconcile position: "
                f"position ID {report.venue_position_id} "
                f"net qty {position.net_qty} != reported {report.net_qty}. "
                f"{report}.",
            )
            return False  # Failed

        return True  # Reconciled

    cdef bint _reconcile_position_report_netting(self, PositionStatusReport report) except *:
        if self.reconcile_by_fake_fills:
            self._generate_fake_fills(report)

        cdef list positions_open = self._cache.positions_open(
            venue=None,  # Faster query filtering
            instrument_id=report.instrument_id,
        )
        net_qty = 0.0
        for position in positions_open:
            net_qty += position.net_qty
        if net_qty != report.net_qty:
            self._log.error(
                f"Cannot reconcile position: "
                f"{report.instrument_id} "
                f"net qty {net_qty} != reported {report.net_qty}.",
            )
            return False  # Failed

        return True  # Reconciled

    cdef void _generate_fake_fills(self, PositionStatusReport report) except *:
        self._log.warning(f"Generating fake fills to pass reconcile check.")

        ts = self._clock.timestamp_ns()
        fake_venue_order_id = VenueOrderId(f"FAKE_ORDER-{ts}")
        # Require exec client generate
        #   nacre.execution.reports.PositionStatusReport
        #   instead of
        #   nautilus_trader.execution.reports.PositionStatusReport
        entry_price = report.entry_px

        order_report = OrderStatusReport(
            account_id=report.account_id,
            instrument_id=report.instrument_id,
            client_order_id=None,
            venue_order_id=fake_venue_order_id,
            order_side=OrderSide.BUY if report.position_side == PositionSide.LONG else OrderSide.SELL,
            order_type=OrderType.LIMIT,
            time_in_force=TimeInForce.GTC,
            order_status=OrderStatus.FILLED,
            price=entry_price,
            quantity=report.quantity,
            filled_qty=report.quantity,
            avg_px=entry_price,
            post_only=False,
            reduce_only=False,
            report_id=UUID4(),
            ts_accepted=ts,
            ts_last=ts,
            ts_init=ts,
            trigger_price=None,
        )

        trade = TradeReport(
            account_id=order_report.account_id,
            instrument_id=order_report.instrument_id,
            venue_order_id=order_report.venue_order_id,
            venue_position_id=report.venue_position_id,
            trade_id=TradeId(UUID4().value),
            order_side=order_report.order_side,
            last_qty=order_report.quantity,
            last_px=entry_price,
            commission=Money(0, Currency.from_str("USDT")),  # HARDCODE for now
            liquidity_side=LiquiditySide.TAKER,
            report_id=UUID4(),
            ts_event=ts,
            ts_init=ts,
        )

        success = self._reconcile_order_report(order_report, [trade])
        self._log.info(f"Generated fake fills {success}")



from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Histogram


WEBSOCKET_DISCONNECT_COUNTER = Counter(
    "nacre_ws_disconnect", "websocket disconnect count", ["endpoint"]
)
WEBSOCKET_RECONNECT_BACKOFF = Gauge(
    "nacre_ws_reconnect_backoff", "websocket reconnect backoff seconds", ["endpoint"]
)
WEBSOCKET_RECV_ERROR_COUNTER = Counter(
    "nacre_ws_errors", "error count in handle websocket recv loop", ["endpoint"]
)


REQ_TIME = Histogram(
    "nacre_http_request_duration_seconds", "time spent in http requests", ["method", "endpoint"]
)
HTTP_ERROR_COUNTER = Counter(
    "nacre_http_request_errors", "error count for http requests", ["method", "endpoint"]
)




# -------------------------------------------------------------------------------------------------
#  Copyright (C) 2015-2022 Nautech Systems Pty Ltd. All rights reserved.
#  https://nautechsystems.io
#
#  Licensed under the GNU Lesser General Public License Version 3.0 (the "License");
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at https://www.gnu.org/licenses/lgpl-3.0.en.html
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# -------------------------------------------------------------------------------------------------
import asyncio
import concurrent.futures
import os
import pathlib
import platform
import signal
import socket
import warnings
from asyncio import AbstractEventLoop
from typing import Callable, Dict, List, Optional, Union

import fsspec
from nautilus_trader.common import Environment
from nautilus_trader.config import ActorFactory
from nautilus_trader.config import CacheConfig
from nautilus_trader.config import CacheDatabaseConfig
from nautilus_trader.config import DataEngineConfig
from nautilus_trader.config import ExecEngineConfig
from nautilus_trader.config import ImportableActorConfig
from nautilus_trader.config import ImportableStrategyConfig
from nautilus_trader.config import LiveDataEngineConfig
from nautilus_trader.config import LiveExecEngineConfig
from nautilus_trader.config import LiveRiskEngineConfig
from nautilus_trader.config import RiskEngineConfig
from nautilus_trader.config import StrategyFactory
from nautilus_trader.config import StreamingConfig
from nautilus_trader.persistence.catalog import resolve_path
from nautilus_trader.persistence.streaming import StreamingFeatherWriter

from nautilus_trader.cache.cache cimport Cache
from nautilus_trader.common.actor cimport Actor
from nautilus_trader.common.clock cimport LiveClock
from nautilus_trader.common.clock cimport TestClock
from nautilus_trader.common.logging cimport LiveLogger
from nautilus_trader.common.logging cimport Logger
from nautilus_trader.common.logging cimport LoggerAdapter
from nautilus_trader.common.logging cimport LogLevel
from nautilus_trader.common.logging cimport nautilus_header
from nautilus_trader.core.correctness cimport Condition
from nautilus_trader.core.datetime cimport nanos_to_millis
from nautilus_trader.core.rust.core cimport unix_timestamp_ns
from nautilus_trader.data.engine cimport DataEngine
from nautilus_trader.execution.engine cimport ExecutionEngine
from nautilus_trader.infrastructure.cache cimport RedisCacheDatabase
from nautilus_trader.live.data_engine cimport LiveDataEngine

# from nautilus_trader.live.execution_engine cimport LiveExecutionEngine
from nautilus_trader.live.risk_engine cimport LiveRiskEngine
from nautilus_trader.msgbus.bus cimport MessageBus
from nautilus_trader.portfolio.portfolio cimport Portfolio
from nautilus_trader.risk.engine cimport RiskEngine
from nautilus_trader.serialization.msgpack.serializer cimport MsgPackSerializer
from nautilus_trader.trading.strategy cimport Strategy
from nautilus_trader.trading.trader cimport Trader

from nacre.live.execution_engine cimport LiveExecutionEngine


try:
    import uvloop
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:  # pragma: no cover
    uvloop = None

from nautilus_trader.system.kernel cimport NautilusKernel as Kernel


cdef class NautilusKernel(Kernel):
    """
    Provides the core Nautilus system kernel

    The kernel is common between backtest, sandbox and live system types.

    Parameters
    ----------
    environment : Environment { ``BACKTEST``, ``SANDBOX``, ``LIVE`` }
        The environment context for the kernel.
    name : str
        The name for the kernel (will prepend all log messages).
    trader_id : str
        The trader ID for the kernel (must be a name and ID tag separated by a hyphen).
    cache_config : CacheConfig
        The cache configuration for the kernel.
    cache_database_config : CacheDatabaseConfig
        The cache database configuration for the kernel.
    data_config : Union[DataEngineConfig, LiveDataEngineConfig]
        The live data engine configuration for the kernel.
    risk_config : Union[RiskEngineConfig, LiveRiskEngineConfig]
        The risk engine configuration for the kernel.
    exec_config : Union[ExecEngineConfig, LiveExecEngineConfig]
        The execution engine configuration for the kernel.
    streaming_config : StreamingConfig, optional
        The configuration for streaming to feather files.
    actor_configs : List[ImportableActorConfig], optional
        The list of importable actor configs.
    strategy_configs : List[ImportableStrategyConfig], optional
        The list of importable strategy configs.
    loop : AbstractEventLoop, optional
        The event loop for the kernel.
    loop_sig_callback : Callable, optional
        The callback for the signal handler.
    loop_debug : bool, default False
        If the event loop should run in debug mode.
    load_state : bool, default False
        If strategy state should be loaded on start.
    save_state : bool, default False
        If strategy state should be saved on stop.
    log_level : LogLevel, default LogLevel.INFO
        The log level for the kernels logger.
    bypass_logging : bool, default False
        If logging to stdout should be bypassed.

    Raises
    ------
    TypeError
        If `environment` is not of type `Environment`.
    ValueError
        If `name` is not a valid string.
    TypeError
        If any configuration object is not of the expected type.

    """

    def __init__(
        self,
        environment not None: Environment,
        str name not None,
        TraderId trader_id not None,
        cache_config not None: CacheConfig,
        cache_database_config not None: CacheDatabaseConfig,
        data_config not None: Union[DataEngineConfig, LiveDataEngineConfig],
        risk_config not None: Union[RiskEngineConfig, LiveRiskEngineConfig],
        exec_config not None: Union[ExecEngineConfig, LiveExecEngineConfig],
        streaming_config: Optional[StreamingConfig] = None,
        actor_configs: Optional[List[ImportableActorConfig]] = None,
        strategy_configs: Optional[List[ImportableStrategyConfig]] = None,
        loop: Optional[AbstractEventLoop] = None,
        loop_sig_callback: Optional[Callable] = None,
        loop_debug: bool = False,
        load_state: bool = False,
        save_state: bool = False,
        LogLevel log_level = LogLevel.INFO,
        bypass_logging: bool = False,
    ):
        if uvloop is None:
            warnings.warn("uvloop is not available.")
        if actor_configs is None:
            actor_configs = []
        if strategy_configs is None:
            strategy_configs = []
        Condition.type(environment, Environment, "environment")
        Condition.valid_string(name, "name")
        Condition.type(cache_config, CacheConfig, "cache_config")
        Condition.type(cache_database_config, CacheDatabaseConfig, "cache_database_config")
        Condition.true(isinstance(data_config, (DataEngineConfig, LiveDataEngineConfig)), "data_config was unrecognized type", ex_type=TypeError)
        Condition.true(isinstance(risk_config, (RiskEngineConfig, LiveRiskEngineConfig)), "risk_config was unrecognized type", ex_type=TypeError)
        Condition.true(isinstance(exec_config, (ExecEngineConfig, LiveExecEngineConfig)), "exec_config was unrecognized type", ex_type=TypeError)
        Condition.type_or_none(streaming_config, StreamingConfig, "streaming_config")

        self.environment = environment

        # Identifiers
        self.name = name
        self.trader_id = trader_id
        self.machine_id = socket.gethostname()
        self.instance_id = UUID4()
        self.ts_created = unix_timestamp_ns()

        # Components
        if self.environment == Environment.BACKTEST:
            self.clock = TestClock()
            self.logger = Logger(
                clock=LiveClock(loop=loop),
                trader_id=self.trader_id,
                machine_id=self.machine_id,
                instance_id=self.instance_id,
                level_stdout=log_level,
                bypass=bypass_logging,
            )
        elif self.environment in (Environment.SANDBOX, Environment.LIVE):
            self.clock = LiveClock(loop=loop)
            self.logger = LiveLogger(
                loop=loop,
                clock=self.clock,
                trader_id=self.trader_id,
                machine_id=self.machine_id,
                instance_id=self.instance_id,
                level_stdout=log_level,
            )
        else:  # pragma: no cover (design-time error)
            raise NotImplementedError(f"environment {environment} not recognized")

        # Setup logging
        self.log = LoggerAdapter(
            component_name=name,
            logger=self.logger,
        )

        nautilus_header(self.log)
        self.log.info("Building system kernel...")

        # Setup loop
        self.loop = loop
        if self.loop is not None:
            self.executor = concurrent.futures.ThreadPoolExecutor()
            self.loop.set_default_executor(self.executor)
            self.loop.set_debug(loop_debug)
            self.loop_sig_callback = loop_sig_callback
            if platform.system() != "Windows":
                # Windows does not support signal handling
                # https://stackoverflow.com/questions/45987985/asyncio-loops-add-signal-handler-in-windows
                self._setup_loop()

        if cache_database_config is None or cache_database_config.type == "in-memory":
            cache_db = None
        elif cache_database_config.type == "redis":
            cache_db = RedisCacheDatabase(
                trader_id=self.trader_id,
                logger=self.logger,
                serializer=MsgPackSerializer(timestamps_as_str=True),
                config=cache_database_config,
            )
        else:
            raise ValueError(
                f"The `cache_db_config.type` is unrecognized. "
                f"Please use one of {{\'in-memory\', \'redis\'}}.",
            )

        ########################################################################
        # Core components
        ########################################################################
        self.msgbus = MessageBus(
            trader_id=self.trader_id,
            clock=self.clock,
            logger=self.logger,
        )

        self.cache = Cache(
            database=cache_db,
            logger=self.logger,
            config=cache_config,
        )

        self.portfolio = Portfolio(
            msgbus=self.msgbus,
            cache=self.cache,
            clock=self.clock,
            logger=self.logger,
        )

        ########################################################################
        # Data Engine
        ########################################################################
        if isinstance(data_config, LiveDataEngineConfig):
            self.data_engine = LiveDataEngine(
                loop=loop,
                msgbus=self.msgbus,
                cache=self.cache,
                clock=self.clock,
                logger=self.logger,
                config=data_config,
            )
        elif isinstance(data_config, DataEngineConfig):
            self.data_engine = DataEngine(
                msgbus=self.msgbus,
                cache=self.cache,
                clock=self.clock,
                logger=self.logger,
                config=data_config,
            )

        ########################################################################
        # Risk Engine
        ########################################################################
        if isinstance(risk_config, LiveRiskEngineConfig):
            self.risk_engine = LiveRiskEngine(
                loop=loop,
                portfolio=self.portfolio,
                msgbus=self.msgbus,
                cache=self.cache,
                clock=self.clock,
                logger=self.logger,
                config=risk_config,
            )
        elif isinstance(risk_config, RiskEngineConfig):
            self.risk_engine = RiskEngine(
                portfolio=self.portfolio,
                msgbus=self.msgbus,
                cache=self.cache,
                clock=self.clock,
                logger=self.logger,
                config=risk_config,
            )

        ########################################################################
        # Execution Engine
        ########################################################################
        if isinstance(exec_config, LiveExecEngineConfig):
            self.exec_engine = LiveExecutionEngine(
                loop=loop,
                msgbus=self.msgbus,
                cache=self.cache,
                clock=self.clock,
                logger=self.logger,
                config=exec_config,
            )
        elif isinstance(exec_config, ExecEngineConfig):
            self.exec_engine = ExecutionEngine(
                msgbus=self.msgbus,
                cache=self.cache,
                clock=self.clock,
                logger=self.logger,
                config=exec_config,
            )

        if exec_config.load_cache:
            self.exec_engine.load_cache()

        ########################################################################
        # Trader
        ########################################################################
        self.trader = Trader(
            trader_id=self.trader_id,
            msgbus=self.msgbus,
            cache=self.cache,
            portfolio=self.portfolio,
            data_engine=self.data_engine,
            risk_engine=self.risk_engine,
            exec_engine=self.exec_engine,
            clock=self.clock,
            logger=self.logger,
        )

        if load_state:
            self.trader.load()

        # Setup writer
        self.writer: Optional[StreamingFeatherWriter] = None
        if streaming_config:
            self._setup_streaming(config=streaming_config)

        # Create importable actors
        for config in actor_configs:
            actor: Actor = ActorFactory.create(config)
            self.trader.add_actor(actor)

        # Create importable strategies
        for config in strategy_configs:
            strategy: Strategy = StrategyFactory.create(config)
            self.trader.add_strategy(strategy)

        cdef int64_t build_time_ms = nanos_to_millis(unix_timestamp_ns() - self.ts_created)
        self.log.info(f"Initialized in {build_time_ms}ms.")
